@@ -1,13 +1,9 @@
 #include "rar.hpp"
 
-static File *CreatedFiles[256];
-static int RemoveCreatedActive=0;
-
 File::File()
 {
   hFile=BAD_HANDLE;
   *FileName=0;
-  *FileNameW=0;
   NewFile=false;
   LastWrite=false;
   HandleType=FILE_HANDLENORMAL;
@@ -16,10 +12,10 @@ File::File()
   ErrorType=FILE_SUCCESS;
   OpenShared=false;
   AllowDelete=true;
-  CloseCount=0;
   AllowExceptions=true;
 #ifdef _WIN_ALL
   NoSequentialRead=false;
+  CreateMode=FMF_UNDEFINED;
 #endif
 }
 
@@ -37,7 +33,6 @@ File::~File()
 void File::operator = (File &SrcFile)
 {
   hFile=SrcFile.hFile;
-  strcpy(FileName,SrcFile.FileName);
   NewFile=SrcFile.NewFile;
   LastWrite=SrcFile.LastWrite;
   HandleType=SrcFile.HandleType;
@@ -45,54 +40,57 @@ void File::operator = (File &SrcFile)
 }
 
 
-bool File::Open(const char *Name,const wchar *NameW,bool OpenShared,bool Update)
+bool File::Open(const wchar *Name,uint Mode)
 {
   ErrorType=FILE_SUCCESS;
   FileHandle hNewFile;
-  if (File::OpenShared)
-    OpenShared=true;
+  bool OpenShared=File::OpenShared || (Mode & FMF_OPENSHARED)!=0;
+  bool UpdateMode=(Mode & FMF_UPDATE)!=0;
+  bool WriteMode=(Mode & FMF_WRITE)!=0;
 #ifdef _WIN_ALL
-  uint Access=GENERIC_READ;
-  if (Update)
+  uint Access=WriteMode ? GENERIC_WRITE:GENERIC_READ;
+  if (UpdateMode)
     Access|=GENERIC_WRITE;
   uint ShareMode=FILE_SHARE_READ;
   if (OpenShared)
     ShareMode|=FILE_SHARE_WRITE;
   uint Flags=NoSequentialRead ? 0:FILE_FLAG_SEQUENTIAL_SCAN;
-  if (WinNT() && NameW!=NULL && *NameW!=0)
-    hNewFile=CreateFileW(NameW,Access,ShareMode,NULL,OPEN_EXISTING,Flags,NULL);
-  else
-    hNewFile=CreateFileA(Name,Access,ShareMode,NULL,OPEN_EXISTING,Flags,NULL);
+  hNewFile=CreateFile(Name,Access,ShareMode,NULL,OPEN_EXISTING,Flags,NULL);
+
+  if (hNewFile==BAD_HANDLE)
+  {
+    wchar LongName[NM];
+    if (GetWinLongPath(Name,LongName,ASIZE(LongName)))
+      hNewFile=CreateFile(LongName,Access,ShareMode,NULL,OPEN_EXISTING,Flags,NULL);
+  }
 
   if (hNewFile==BAD_HANDLE && GetLastError()==ERROR_FILE_NOT_FOUND)
     ErrorType=FILE_NOTFOUND;
 #else
-  int flags=Update ? O_RDWR:O_RDONLY;
+  int flags=UpdateMode ? O_RDWR:(WriteMode ? O_WRONLY:O_RDONLY);
 #ifdef O_BINARY
   flags|=O_BINARY;
 #if defined(_AIX) && defined(_LARGE_FILE_API)
   flags|=O_LARGEFILE;
 #endif
 #endif
-#if defined(_EMX) && !defined(_DJGPP)
-  int sflags=OpenShared ? SH_DENYNO:SH_DENYWR;
-  int handle=sopen(Name,flags,sflags);
-#else
-  int handle=open(Name,flags);
+  char NameA[NM];
+  WideToChar(Name,NameA,ASIZE(NameA));
+
+  int handle=open(NameA,flags);
 #ifdef LOCK_EX
 
 #ifdef _OSF_SOURCE
   extern "C" int flock(int, int);
 #endif
 
-  if (!OpenShared && Update && handle>=0 && flock(handle,LOCK_EX|LOCK_NB)==-1)
+  if (!OpenShared && UpdateMode && handle>=0 && flock(handle,LOCK_EX|LOCK_NB)==-1)
   {
     close(handle);
-    return(false);
+    return false;
   }
 #endif
-#endif
-  hNewFile=handle==-1 ? BAD_HANDLE:fdopen(handle,Update ? UPDATEBINARY:READBINARY);
+  hNewFile=handle==-1 ? BAD_HANDLE:fdopen(handle,UpdateMode ? UPDATEBINARY:READBINARY);
   if (hNewFile==BAD_HANDLE && errno==ENOENT)
     ErrorType=FILE_NOTFOUND;
 #endif
@@ -103,100 +101,80 @@ bool File::Open(const char *Name,const wchar *NameW,bool OpenShared,bool Update)
   if (Success)
   {
     hFile=hNewFile;
-
-    // We use memove instead of strcpy and wcscpy to avoid problems
-    // with overlapped buffers. While we do not call this function with
-    // really overlapped buffers yet, we do call it with Name equal to
-    // FileName like Arc.Open(Arc.FileName,Arc.FileNameW).
-    if (NameW!=NULL)
-      memmove(FileNameW,NameW,(wcslen(NameW)+1)*sizeof(*NameW));
-    else
-      *FileNameW=0;
-    if (Name!=NULL)
-      memmove(FileName,Name,strlen(Name)+1);
-    else
-      WideToChar(NameW,FileName);
-    AddFileToList(hFile);
+    wcsncpyz(FileName,Name,ASIZE(FileName));
   }
-  return(Success);
+  return Success;
 }
 
 
 #if !defined(SHELL_EXT) && !defined(SFX_MODULE)
-void File::TOpen(const char *Name,const wchar *NameW)
+void File::TOpen(const wchar *Name)
 {
-  if (!WOpen(Name,NameW))
-    ErrHandler.Exit(OPEN_ERROR);
+  if (!WOpen(Name))
+    ErrHandler.Exit(RARX_OPEN);
 }
 #endif
 
 
-bool File::WOpen(const char *Name,const wchar *NameW)
+bool File::WOpen(const wchar *Name)
 {
-  if (Open(Name,NameW))
-    return(true);
-  ErrHandler.OpenErrorMsg(Name,NameW);
-  return(false);
+  if (Open(Name))
+    return true;
+  ErrHandler.OpenErrorMsg(Name);
+  return false;
 }
 
 
-bool File::Create(const char *Name,const wchar *NameW,bool ShareRead)
+bool File::Create(const wchar *Name,uint Mode)
 {
+  // OpenIndiana based NAS and CIFS shares fail to set the file time if file
+  // was created in read+write mode and some data was written and not flushed
+  // before SetFileTime call. So we should use the write only mode if we plan
+  // SetFileTime call and do not need to read from file.
+  bool WriteMode=(Mode & FMF_WRITE)!=0;
+  bool ShareRead=(Mode & FMF_SHAREREAD)!=0 || File::OpenShared;
 #ifdef _WIN_ALL
-  DWORD ShareMode=(ShareRead || File::OpenShared) ? FILE_SHARE_READ:0;
-  if (WinNT() && NameW!=NULL && *NameW!=0)
-    hFile=CreateFileW(NameW,GENERIC_READ|GENERIC_WRITE,ShareMode,NULL,
-                      CREATE_ALWAYS,0,NULL);
-  else
-    hFile=CreateFileA(Name,GENERIC_READ|GENERIC_WRITE,ShareMode,NULL,
-                     CREATE_ALWAYS,0,NULL);
+  CreateMode=Mode;
+  uint Access=WriteMode ? GENERIC_WRITE:GENERIC_READ|GENERIC_WRITE;
+  DWORD ShareMode=ShareRead ? FILE_SHARE_READ:0;
+  hFile=CreateFile(Name,Access,ShareMode,NULL,CREATE_ALWAYS,0,NULL);
+
+  if (hFile==BAD_HANDLE)
+  {
+    wchar LongName[NM];
+    if (GetWinLongPath(Name,LongName,ASIZE(LongName)))
+      hFile=CreateFile(LongName,Access,ShareMode,NULL,CREATE_ALWAYS,0,NULL);
+  }
+
 #else
-  hFile=fopen(Name,CREATEBINARY);
+  char NameA[NM];
+  WideToChar(Name,NameA,ASIZE(NameA));
+  hFile=fopen(NameA,WriteMode ? WRITEBINARY:CREATEBINARY);
 #endif
   NewFile=true;
   HandleType=FILE_HANDLENORMAL;
   SkipClose=false;
-  if (NameW!=NULL)
-    wcscpy(FileNameW,NameW);
-  else
-    *FileNameW=0;
-  if (Name!=NULL)
-    strcpy(FileName,Name);
-  else
-    WideToChar(NameW,FileName);
-  AddFileToList(hFile);
-  return(hFile!=BAD_HANDLE);
-}
-
-
-void File::AddFileToList(FileHandle hFile)
-{
-  if (hFile!=BAD_HANDLE)
-    for (int I=0;I<sizeof(CreatedFiles)/sizeof(CreatedFiles[0]);I++)
-      if (CreatedFiles[I]==NULL)
-      {
-        CreatedFiles[I]=this;
-        break;
-      }
+  wcsncpyz(FileName,Name,ASIZE(FileName));
+  return hFile!=BAD_HANDLE;
 }
 
 
 #if !defined(SHELL_EXT) && !defined(SFX_MODULE)
-void File::TCreate(const char *Name,const wchar *NameW,bool ShareRead)
+void File::TCreate(const wchar *Name,uint Mode)
 {
-  if (!WCreate(Name,NameW,ShareRead))
-    ErrHandler.Exit(FATAL_ERROR);
+  if (!WCreate(Name,Mode))
+    ErrHandler.Exit(RARX_FATAL);
 }
 #endif
 
 
-bool File::WCreate(const char *Name,const wchar *NameW,bool ShareRead)
+bool File::WCreate(const wchar *Name,uint Mode)
 {
-  if (Create(Name,NameW,ShareRead))
-    return(true);
-  ErrHandler.SetErrorCode(CREATE_ERROR);
-  ErrHandler.CreateErrorMsg(Name,NameW);
-  return(false);
+  if (Create(Name,Mode))
+    return true;
+  ErrHandler.SetErrorCode(RARX_CREATE);
+  ErrHandler.CreateErrorMsg(Name);
+  return false;
 }
 
 
@@ -215,20 +193,12 @@ bool File::Close()
 #else
         Success=fclose(hFile)!=EOF;
 #endif
-        if (Success || !RemoveCreatedActive)
-          for (int I=0;I<sizeof(CreatedFiles)/sizeof(CreatedFiles[0]);I++)
-            if (CreatedFiles[I]==this)
-            {
-              CreatedFiles[I]=NULL;
-              break;
-            }
       }
       hFile=BAD_HANDLE;
       if (!Success && AllowExceptions)
-        ErrHandler.CloseError(FileName,FileNameW);
+        ErrHandler.CloseError(FileName);
     }
-  CloseCount++;
-  return(Success);
+  return Success;
 }
 
 
@@ -245,32 +215,27 @@ void File::Flush()
 bool File::Delete()
 {
   if (HandleType!=FILE_HANDLENORMAL)
-    return(false);
+    return false;
   if (hFile!=BAD_HANDLE)
     Close();
   if (!AllowDelete)
-    return(false);
-  return(DelFile(FileName,FileNameW));
+    return false;
+  return DelFile(FileName);
 }
 
 
-bool File::Rename(const char *NewName,const wchar *NewNameW)
+bool File::Rename(const wchar *NewName)
 {
-  // we do not need to rename if names are already same
-  bool Success=strcmp(FileName,NewName)==0;
-  if (Success && *FileNameW!=0 && *NullToEmpty(NewNameW)!=0)
-    Success=wcscmp(FileNameW,NewNameW)==0;
+  // No need to rename if names are already same.
+  bool Success=wcscmp(FileName,NewName)==0;
 
   if (!Success)
-    Success=RenameFile(FileName,FileNameW,NewName,NewNameW);
+    Success=RenameFile(FileName,NewName);
 
   if (Success)
-  {
-    // renamed successfully, storing the new name
-    strcpy(FileName,NewName);
-    wcscpy(FileNameW,NullToEmpty(NewNameW));
-  }
-  return(Success);
+    wcscpy(FileName,NewName);
+
+  return Success;
 }
 
 
@@ -328,9 +293,9 @@ void File::Write(const void *Data,size_t Size)
       uint64 FreeSize=GetFreeDisk(FileName);
       SetLastError(ErrCode);
       if (FreeSize>Size && FilePos-Size<=0xffffffff && FilePos+Size>0xffffffff)
-        ErrHandler.WriteErrorFAT(FileName,FileNameW);
+        ErrHandler.WriteErrorFAT(FileName);
 #endif
-      if (ErrHandler.AskRepeatWrite(FileName,FileNameW,false))
+      if (ErrHandler.AskRepeatWrite(FileName,false))
       {
 #ifndef _WIN_ALL
         clearerr(hFile);
@@ -339,7 +304,7 @@ void File::Write(const void *Data,size_t Size)
           Seek(Tell()-Written,SEEK_SET);
         continue;
       }
-      ErrHandler.WriteError(NULL,NULL,FileName,FileNameW);
+      ErrHandler.WriteError(NULL,FileName);
     }
     break;
   }
@@ -374,14 +339,14 @@ int File::Read(void *Data,size_t Size)
         }
         else
         {
-          if (HandleType==FILE_HANDLENORMAL && ErrHandler.AskRepeatRead(FileName,FileNameW))
+          if (HandleType==FILE_HANDLENORMAL && ErrHandler.AskRepeatRead(FileName))
             continue;
-          ErrHandler.ReadError(FileName,FileNameW);
+          ErrHandler.ReadError(FileName);
         }
     }
     break;
   }
-  return(ReadSize);
+  return ReadSize;
 }
 
 
@@ -390,6 +355,7 @@ int File::DirectRead(void *Data,size_t Size)
 {
 #ifdef _WIN_ALL
   const size_t MaxDeviceRead=20000;
+  const size_t MaxLockedRead=32768;
 #endif
 #ifndef _WIN_CE
   if (HandleType==FILE_HANDLESTD)
@@ -408,12 +374,23 @@ int File::DirectRead(void *Data,size_t Size)
   if (!ReadFile(hFile,Data,(DWORD)Size,&Read,NULL))
   {
     if (IsDevice() && Size>MaxDeviceRead)
-      return(DirectRead(Data,MaxDeviceRead));
+      return DirectRead(Data,MaxDeviceRead);
     if (HandleType==FILE_HANDLESTD && GetLastError()==ERROR_BROKEN_PIPE)
-      return(0);
-    return(-1);
+      return 0;
+
+    // We had a bug report about failure to archive 1C database lock file
+    // 1Cv8tmp.1CL, which is a zero length file with a region above 200 KB
+    // permanently locked. If our first read request uses too large buffer
+    // and if we are in -dh mode, so we were able to open the file,
+    // we'll fail with "Read error". So now we use try a smaller buffer size
+    // in case of lock error.
+    if (HandleType==FILE_HANDLENORMAL && Size>MaxLockedRead &&
+        GetLastError()==ERROR_LOCK_VIOLATION)
+      return DirectRead(Data,MaxLockedRead);
+
+    return -1;
   }
-  return(Read);
+  return Read;
 #else
   if (LastWrite)
   {
@@ -423,8 +400,8 @@ int File::DirectRead(void *Data,size_t Size)
   clearerr(hFile);
   size_t ReadSize=fread(Data,1,Size,hFile);
   if (ferror(hFile))
-    return(-1);
-  return((int)ReadSize);
+    return -1;
+  return (int)ReadSize;
 #endif
 }
 
@@ -432,14 +409,14 @@ int File::DirectRead(void *Data,size_t Size)
 void File::Seek(int64 Offset,int Method)
 {
   if (!RawSeek(Offset,Method) && AllowExceptions)
-    ErrHandler.SeekError(FileName,FileNameW);
+    ErrHandler.SeekError(FileName);
 }
 
 
 bool File::RawSeek(int64 Offset,int Method)
 {
   if (hFile==BAD_HANDLE)
-    return(true);
+    return true;
   if (Offset<0 && Method!=SEEK_SET)
   {
     Offset=(Method==SEEK_CUR ? Tell():FileLength())+Offset;
@@ -449,7 +426,7 @@ bool File::RawSeek(int64 Offset,int Method)
   LONG HighDist=(LONG)(Offset>>32);
   if (SetFilePointer(hFile,(LONG)Offset,&HighDist,Method)==0xffffffff &&
       GetLastError()!=NO_ERROR)
-    return(false);
+    return false;
 #else
   LastWrite=false;
 #if defined(_LARGEFILE_SOURCE) && !defined(_OSF_SOURCE) && !defined(__VMS)
@@ -457,9 +434,9 @@ bool File::RawSeek(int64 Offset,int Method)
 #else
   if (fseek(hFile,(long)Offset,Method)!=0)
 #endif
-    return(false);
+    return false;
 #endif
-  return(true);
+  return true;
 }
 
 
@@ -467,23 +444,23 @@ int64 File::Tell()
 {
   if (hFile==BAD_HANDLE)
     if (AllowExceptions)
-      ErrHandler.SeekError(FileName,FileNameW);
+      ErrHandler.SeekError(FileName);
     else
-      return(-1);
+      return -1;
 #ifdef _WIN_ALL
   LONG HighDist=0;
   uint LowDist=SetFilePointer(hFile,0,&HighDist,FILE_CURRENT);
   if (LowDist==0xffffffff && GetLastError()!=NO_ERROR)
     if (AllowExceptions)
-      ErrHandler.SeekError(FileName,FileNameW);
+      ErrHandler.SeekError(FileName);
     else
-      return(-1);
-  return(INT32TO64(HighDist,LowDist));
+      return -1;
+  return INT32TO64(HighDist,LowDist);
 #else
 #if defined(_LARGEFILE_SOURCE) && !defined(_OSF_SOURCE)
-  return(ftello(hFile));
+  return ftello(hFile);
 #else
-  return(ftell(hFile));
+  return ftell(hFile);
 #endif
 #endif
 }
@@ -498,6 +475,14 @@ void File::Prealloc(int64 Size)
     Seek(0,SEEK_SET);
   }
 #endif
+
+#if defined(_UNIX) && defined(USE_FALLOCATE)
+  // fallocate is rather new call. Only latest kernels support it.
+  // So we are not using it by default yet.
+  int fd = fileno(hFile);
+  if (fd >= 0)
+    fallocate(fd, 0, 0, Size);
+#endif
 }
 
 
@@ -505,7 +490,7 @@ byte File::GetByte()
 {
   byte Byte=0;
   Read(&Byte,1);
-  return(Byte);
+  return Byte;
 }
 
 
@@ -518,9 +503,9 @@ void File::PutByte(byte Byte)
 bool File::Truncate()
 {
 #ifdef _WIN_ALL
-  return(SetEndOfFile(hFile)==TRUE);
+  return SetEndOfFile(hFile)==TRUE;
 #else
-  return(false);
+  return false;
 #endif
 }
 
@@ -528,6 +513,12 @@ bool File::Truncate()
 void File::SetOpenFileTime(RarTime *ftm,RarTime *ftc,RarTime *fta)
 {
 #ifdef _WIN_ALL
+  // Workaround for OpenIndiana NAS time bug. If we cannot create a file
+  // in write only mode, we need to flush the write buffer before calling
+  // SetFileTime or file time will not be changed.
+  if (CreateMode!=FMF_UNDEFINED && (CreateMode & FMF_WRITE)==0)
+    FlushFileBuffers(hFile);
+
   bool sm=ftm!=NULL && ftm->IsSet();
   bool sc=ftc!=NULL && ftc->IsSet();
   bool sa=fta!=NULL && fta->IsSet();
@@ -545,15 +536,15 @@ void File::SetOpenFileTime(RarTime *ftm,RarTime *ftc,RarTime *fta)
 
 void File::SetCloseFileTime(RarTime *ftm,RarTime *fta)
 {
-#if defined(_UNIX) || defined(_EMX)
+#ifdef _UNIX
   SetCloseFileTimeByName(FileName,ftm,fta);
 #endif
 }
 
 
-void File::SetCloseFileTimeByName(const char *Name,RarTime *ftm,RarTime *fta)
+void File::SetCloseFileTimeByName(const wchar *Name,RarTime *ftm,RarTime *fta)
 {
-#if defined(_UNIX) || defined(_EMX)
+#ifdef _UNIX
   bool setm=ftm!=NULL && ftm->IsSet();
   bool seta=fta!=NULL && fta->IsSet();
   if (setm || seta)
@@ -567,7 +558,9 @@ void File::SetCloseFileTimeByName(const char *Name,RarTime *ftm,RarTime *fta)
       ut.actime=fta->GetUnix();
     else
       ut.actime=ut.modtime;
-    utime(Name,&ut);
+    char NameA[NM];
+    WideToChar(Name,NameA,ASIZE(NameA));
+    utime(NameA,&ut);
   }
 #endif
 }
@@ -592,78 +585,23 @@ int64 File::FileLength()
 {
   SaveFilePos SavePos(*this);
   Seek(0,SEEK_END);
-  return(Tell());
-}
-
-
-void File::SetHandleType(FILE_HANDLETYPE Type)
-{
-  HandleType=Type;
+  return Tell();
 }
 
 
 bool File::IsDevice()
 {
   if (hFile==BAD_HANDLE)
-    return(false);
+    return false;
 #ifdef _WIN_ALL
   uint Type=GetFileType(hFile);
-  return(Type==FILE_TYPE_CHAR || Type==FILE_TYPE_PIPE);
+  return Type==FILE_TYPE_CHAR || Type==FILE_TYPE_PIPE;
 #else
-  return(isatty(fileno(hFile)));
+  return isatty(fileno(hFile));
 #endif
 }
 
 
-#ifndef SFX_MODULE
-void File::fprintf(const char *fmt,...)
-{
-  va_list argptr;
-  va_start(argptr,fmt);
-  safebuf char Msg[2*NM+1024],OutMsg[2*NM+1024];
-  vsprintf(Msg,fmt,argptr);
-#ifdef _WIN_ALL
-  for (int Src=0,Dest=0;;Src++)
-  {
-    char CurChar=Msg[Src];
-    if (CurChar=='\n')
-      OutMsg[Dest++]='\r';
-    OutMsg[Dest++]=CurChar;
-    if (CurChar==0)
-      break;
-  }
-#else
-  strcpy(OutMsg,Msg);
-#endif
-  Write(OutMsg,strlen(OutMsg));
-  va_end(argptr);
-}
-#endif
-
-
-bool File::RemoveCreated()
-{
-  RemoveCreatedActive++;
-  bool RetCode=true;
-  for (int I=0;I<sizeof(CreatedFiles)/sizeof(CreatedFiles[0]);I++)
-    if (CreatedFiles[I]!=NULL)
-    {
-      CreatedFiles[I]->SetExceptions(false);
-      bool Success;
-      if (CreatedFiles[I]->NewFile)
-        Success=CreatedFiles[I]->Delete();
-      else
-        Success=CreatedFiles[I]->Close();
-      if (Success)
-        CreatedFiles[I]=NULL;
-      else
-        RetCode=false;
-    }
-  RemoveCreatedActive--;
-  return(RetCode);
-}
-
-
 #ifndef SFX_MODULE
 int64 File::Copy(File &Dest,int64 Length)
 {
@@ -683,6 +621,6 @@ int64 File::Copy(File &Dest,int64 Length)
     if (!CopyAll)
       Length-=ReadSize;
   }
-  return(CopySize);
+  return CopySize;
 }
 #endif
