@@ -1,10 +1,12 @@
 #ifndef _RAR_OPTIONS_
 #define _RAR_OPTIONS_
 
-#define DEFAULT_RECOVERY    -1
+#define DEFAULT_RECOVERY     -3
 
 #define DEFAULT_RECVOLUMES  -10
 
+#define VOLSIZE_AUTO   INT64NDF // Automatically detect the volume size.
+
 enum PATH_EXCL_MODE {
   EXCL_UNCHANGED=0,    // Process paths as is (default).
   EXCL_SKIPWHOLEPATH,  // -ep  (exclude the path completely)
@@ -40,19 +42,27 @@ enum RECURSE_MODE
 
 enum OVERWRITE_MODE 
 {
-  OVERWRITE_DEFAULT=0, // ask for extraction, silently overwrite for archiving
+  OVERWRITE_DEFAULT=0, // Ask when extracting, silently overwrite when archiving.
   OVERWRITE_ALL,
   OVERWRITE_NONE,
   OVERWRITE_AUTORENAME,
   OVERWRITE_FORCE_ASK
 };
 
+
+enum QOPEN_MODE { QOPEN_NONE, QOPEN_AUTO, QOPEN_ALWAYS };
+
 enum RAR_CHARSET { RCH_DEFAULT=0,RCH_ANSI,RCH_OEM,RCH_UNICODE };
 
 #define     MAX_FILTER_TYPES           16
 enum FilterState {FILTER_DEFAULT=0,FILTER_AUTO,FILTER_FORCE,FILTER_DISABLE};
 
 
+enum SAVECOPY_MODE {
+  SAVECOPY_NONE=0, SAVECOPY_SILENT, SAVECOPY_LIST, SAVECOPY_LISTEXIT,
+  SAVECOPY_DUPLISTEXIT
+};
+
 struct FilterMode
 {
   FilterState State;
@@ -73,24 +83,26 @@ class RAROptions
     uint ExclFileAttr;
     uint InclFileAttr;
     bool InclAttrSet;
-    uint WinSize;
-    char TempPath[NM];
+    size_t WinSize;
+    wchar TempPath[NM];
+#ifdef USE_QOPEN
+    wchar SFXModule[NM];
+    QOPEN_MODE QOpenMode;
+#endif
     bool ConfigDisabled; // Switch -cfg-.
-    char ExtrPath[NM];
-    wchar ExtrPathW[NM];
-    char CommentFile[NM];
-    wchar CommentFileW[NM];
+    wchar ExtrPath[NM];
+    wchar CommentFile[NM];
     RAR_CHARSET CommentCharset;
     RAR_CHARSET FilelistCharset;
-    char ArcPath[NM];
-    wchar ArcPathW[NM];
+    wchar ArcPath[NM];
     SecPassword Password;
     bool EncryptHeaders;
-    char LogName[NM];
+    wchar LogName[NM];
     MESSAGE_TYPE MsgStream;
     bool Sound;
     OVERWRITE_MODE Overwrite;
     int Method;
+    HASH_TYPE HashType;
     int Recovery;
     int RecVolNumber;
     bool DisablePercentage;
@@ -100,7 +112,6 @@ class RAROptions
     int SolidCount;
     bool ClearArc;
     bool AddArcOnly;
-    bool AV;
     bool DisableComment;
     bool FreshFiles;
     bool UpdateFiles;
@@ -110,20 +121,22 @@ class RAROptions
     Array<int64> NextVolSizes;
     uint CurVolNum;
     bool AllYes;
-    bool DisableViewAV;
     bool DisableSortSolid;
     int ArcTime;
     int ConvertNames;
     bool ProcessOwners;
-    bool SaveLinks;
+    bool SaveSymLinks;
+    bool SaveHardLinks;
     int Priority;
     int SleepTime;
     bool KeepBroken;
     bool OpenShared;
     bool DeleteFiles;
+
+
 #ifndef SFX_MODULE
     bool GenerateArcName;
-    char GenerateMask[MAX_GENERATE_MASK];
+    wchar GenerateMask[MAX_GENERATE_MASK];
 #endif
     bool SyncFiles;
     bool ProcessEA;
@@ -134,34 +147,27 @@ class RAROptions
     RarTime FileTimeAfter;
     int64 FileSizeLess;
     int64 FileSizeMore;
-    bool OldNumbering;
     bool Lock;
     bool Test;
     bool VolumePause;
     FilterMode FilterModes[MAX_FILTER_TYPES];
-    char EmailTo[NM];
+    wchar EmailTo[NM];
     uint VersionControl;
-    bool NoEndBlock;
     bool AppendArcNameToPath;
     bool Shutdown;
     EXTTIME_MODE xmtime;
     EXTTIME_MODE xctime;
     EXTTIME_MODE xatime;
-    EXTTIME_MODE xarctime;
-    char CompressStdin[NM];
-
-#ifdef RAR_SMP
-    uint Threads;
-#endif
+    wchar CompressStdin[NM];
 
+    uint Threads; // We use it to init hash even if RAR_SMP is not defined.
 
 
 
 
 
 #ifdef RARDLL
-    char DllDestName[NM];
-    wchar DllDestNameW[NM];
+    wchar DllDestName[NM];
     int DllOpMode;
     int DllError;
     LPARAM UserData;
